/*
 * past_api.h: This file is part of the PAST project.
 *
 * PAST: the PoCC Abstract Syntax Tree
 *
 * Copyright (C) 2011 Louis-Noel Pouchet
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * The complete GNU Lesser General Public Licence Notice can be found
 * as the `COPYING.LESSER' file in the root directory.
 *
 * Author:
 * Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 *
 */
#ifndef PAST_PAST_API_H
# define PAST_PAST_API_H

# include <past/common.h>
# include <past/symbols.h>
# include <past/past.h>

BEGIN_C_DECLS

/**
 * Count the number of for loops in the (sub-)tree dominated by 'root'.
 *
 */
extern
int
past_count_for_loops (s_past_node_t* root);

/**
 * Count the number of statements (past_cloogstmt and past_statement)
 * in a tree.
 *
 */
extern
int
past_count_statements (s_past_node_t* root);

/**
 * Count the number of nodes of type 'type' in a tree.
 *
 */
extern
int
past_count_nodetype (s_past_node_t* root, cs_past_node_type_t* type);

/**
 * Collect all nodes of type 'type' in a tree, prefix order.
 *
 */
extern
s_past_node_t**
past_collect_nodetype (s_past_node_t* root, cs_past_node_type_t* type);

/**
 * Return the maximal loop depth in the (sub-)tree dominated by 'root'.
 *
 */
extern
int
past_max_loop_depth (s_past_node_t* root);

/**
 * Return the loop depth of a node, relative to the whole tree.
 *
 */
extern
int
past_loop_depth (s_past_node_t* root);

/**
 * Return a statement/cloogstmt surrounded by 'depth' loop(s) in the
 * (sub-)tree dominated by 'root'.
 *
 */
extern
s_past_node_t*
past_find_statement_at_depth (s_past_node_t* root, int depth);

/**
 * Return a NULL-terminated array of the outer loops in the (sub-)tree
 * dominated by 'root'.
 *
 */
extern
s_past_node_t**
past_outer_loops (s_past_node_t* root);

/**
 * Return a NULL-terminated array of the inner loops in the (sub-)tree
 * dominated by 'root'.
 *
 */
extern
s_past_node_t**
past_inner_loops (s_past_node_t* root);

/**
 * Return true if the (sub-)tree dominated by 'root' contains a loop.
 *
 */
extern
int
past_contain_loop (s_past_node_t* root);

/**
 * Return true if the node is an outer loop in the whole tree.
 *
 */
extern
int
past_is_outer_for_loop (s_past_node_t* node);

/**
 * Return true if the node is an inner loop in the whole tree.
 *
 */
extern
int
past_is_inner_for_loop (s_past_node_t* node);

/**
 * Rebuild the symbol table, and attach it to the root node.
 *
 */
extern
void
past_rebuild_symbol_table (s_past_node_t* node);

/**
 * Optimize 'for' loop bounds, by hoisting them as much as possible.
 *
 */
extern
void
past_optimize_loop_bounds (s_past_node_t* root);

/**
 * Replace all array references by an API-specified function
 * call. Returns a possibly modified root for the tree.
 *
 * E.g:
 * A[i+42][j] = ... is replaced by *__past_arrayref_write_2d(A, i+42, j) = ...
 *
 */
extern
s_past_node_t*
past_encapsulate_arrayref_in_func (s_past_node_t* root);

/**
 * Return the immediate enclosing node of type 'type', NULL if it does
 * not exists.
 *
 */
extern
s_past_node_t*
past_get_enclosing_node (s_past_node_t* node, cs_past_node_type_t* type);

/**
 * Return a null-terminated set of unique symbols being read in the
 * tree.
 *
 */
extern
s_symbol_t**
past_collect_read_symbols (s_past_node_t* node);

/**
 * Return a null-terminated set of unique symbols being written in the
 * tree.
 *
 */
extern
s_symbol_t**
past_collect_write_symbols (s_past_node_t* node);

/**
 * Return a null-terminated set of unique array symbols in the tree.
 *
 */
extern
s_symbol_t**
past_collect_array_symbols (s_past_node_t* node);

/**
 * Return a null-terminated set of unique symbols being read in any
 * affine expression in the tree. By construction those should be
 * integers.
 *
 */
extern
s_symbol_t**
past_collect_affineexpr_symbols (s_past_node_t* node);

/**
 * Return a null-terminated set of unique loop iterator symbols
 * assigned in the tree.
 *
 */
extern
s_symbol_t**
past_collect_written_loopiterators_symbols (s_past_node_t* node);

/**
 * Replace all inner-most SIMD loop by an API-specified function call.
 *
 * E.g:
 * for (i = 0; i < N; i += 2)
 *   A[k][i] *= alpha;
 * is replaced by:
 * __past_loop_1(&i, N, k, A, alpha)
 *
 * And a new tree is created and contains:
 *
 * void __past_loop_1(int* i_ptr, int N, int k, DATA_TYPE PAST_DECL_A A,
 *                    DATA_TYPE alpha) {
 *  int i = *i_ptr;
 *  for (i = 0; i < N; i += 2)
 *     A[k][i] *= alpha;
 *  *i_ptr = i;
 * }
 *
 */
extern
s_past_node_t*
past_encapsulate_simd_loops_in_func (s_past_node_t** root);

/**
 * Replace all inner-most SIMD loop by an API-specified function call.
 *
 * E.g:
 * for (i = 0; i < N; i += 2)
 *   A[k][i] *= alpha;
 * is replaced by:
 * __past_loop_1(N, k, alpha, A, &i)
 *
 * And a separate file 'filename' is created and contains:
 *
 * void __past_loop_1(int N, int k, DATA_TYPE alpha, DATA_TYPE PAST_DECL_A(A),
 *                    int* __local_i) {
 *  int i = *__local_i;
 *  for (i = 0; i < N; i += 2)
 *     A[k][i] *= alpha;
 *  *__local_i = i;
 * }
 *
 */
extern
void
past_encapsulate_simd_loops_in_func_file (s_past_node_t* root,
					  char* filename);


END_C_DECLS


#endif // PAST_PAST_API_H
