/*
 * past_api.c: This file is part of the PAST project.
 *
 * PAST: the PoCC Abstract Syntax Tree
 *
 * Copyright (C) 2011 Louis-Noel Pouchet
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * The complete GNU Lesser General Public Licence Notice can be found
 * as the `COPYING.LESSER' file in the root directory.
 *
 * Author:
 * Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 *
 */
#if HAVE_CONFIG_H
# include <past/config.h>
#endif

#include <assert.h>
#include <past/common.h>
#include <past/past_api.h>




static
void traverse_count_nodetype (s_past_node_t* node, void* args)
{
  cs_past_node_type_t* type = (cs_past_node_type_t*) ((void**)args)[0];
  if (past_node_is_a (node, type))
    {
      int* count = ((void**)args)[1];
      (*count)++;
    }
}

/**
 * Count the number of nodes of type 'type' in a tree.
 *
 */
int
past_count_nodetype (s_past_node_t* root, cs_past_node_type_t* type)
{
  int num_nodes = 0;
  if (root)
    {
      void* args[2];
      args[0] = (void*)type;
      args[1] = &num_nodes;
      s_past_node_t* next = root->next;
      root->next = NULL;
      past_visitor (root, traverse_count_nodetype, (void*)args, NULL, NULL);
      root->next = next;
    }

  return num_nodes;
}


static
void traverse_collect_nodetype (s_past_node_t* node, void* args)
{
  cs_past_node_type_t* type = (cs_past_node_type_t*) ((void**)args)[0];
  if (past_node_is_a (node, type))
    {
      s_past_node_t** res = ((void**)args)[1];
      int i;
      for (i = 0; res[i]; ++i)
	;
      res[i] = node;
    }
}
/**
 * Collect all nodes of type 'type' in a tree, prefix order.
 *
 */
s_past_node_t**
past_collect_nodetype (s_past_node_t* root, cs_past_node_type_t* type)
{
  if (! root)
    return NULL;
  int num_nodes = past_count_nodetype (root, type);
  s_past_node_t** res = XMALLOC(s_past_node_t*, num_nodes + 1);
  int i;
  for (i = 0; i < num_nodes + 1; ++i)
    res[i] = NULL;
  void* args[2];
  args[0] = (void*)type;
  args[1] = res;
  s_past_node_t* next = root->next;
  root->next = NULL;
  past_visitor (root, traverse_collect_nodetype, (void*)args, NULL, NULL);
  root->next = next;

  return res;
}

/**
 * Count the number of for loops in a tree.
 *
 */
int
past_count_for_loops (s_past_node_t* root)
{
  return past_count_nodetype (root, past_for);
}

/**
 * Count the number of statements (past_cloogstmt and past_statement)
 * in a tree.
 *
 */
int
past_count_statements (s_past_node_t* root)
{
  return past_count_nodetype (root, past_cloogstmt) +
    past_count_nodetype (root, past_statement);
}

static
int
past_loop_depth_sub (s_past_node_t* root, s_past_node_t* top)
{
  if (!root || root == top)
    return 0;

  s_past_node_t* node;
  int depth = 0;
  do
    {
      if (past_node_is_a (root, past_for))
	++depth;
      if (root == top)
	break;
      root = root->parent;
    }
  while (root);

  return depth;
}

static
void traverse_max_loop_depth (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_cloogstmt) ||
      past_node_is_a (node, past_statement))
    {
      s_past_node_t* top = ((void**)data)[0];
      int* maxdepth = ((void**)data)[1];
      int depth = past_loop_depth_sub (node, top);
      if (depth > *maxdepth)
	*maxdepth = depth;
    }
}

/**
 * Return the maximal loop depth.
 *
 */
int
past_max_loop_depth (s_past_node_t* root)
{
  int maxdepth = 0;
  void* args[2];
  args[0] = root;
  args[1] = &maxdepth;
  s_past_node_t* next = root->next;
  root->next = NULL;
  past_visitor (root, traverse_max_loop_depth, (void*) args, NULL, NULL);
  root->next = next;

  return maxdepth;
}


/**
 * Return the loop depth of a node.
 *
 */
int
past_loop_depth (s_past_node_t* root)
{
  return past_loop_depth_sub (root, NULL);
}

static
void traverse_find_statement_at_depth (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_cloogstmt) ||
      past_node_is_a (node, past_statement))
    {
      s_past_node_t* found = ((void**)data)[2];
      if (! found)
	{
	  s_past_node_t* top = ((void**)data)[0];
	  int* maxdepth = ((void**)data)[1];
	  int depth = past_loop_depth_sub (node, top);
	  if (depth == *maxdepth)
	    ((void**)data)[2] = node;
	}
    }
}

/**
 * Return a statement/cloogstmt surrounded by 'depth' loop.
 *
 */
s_past_node_t*
past_find_statement_at_depth (s_past_node_t* root, int depth)
{
  void* args[3];
  args[0] = root;
  args[1] = &depth;
  args[2] = NULL;
  s_past_node_t* next = root->next;
  root->next = NULL;
  past_visitor (root, traverse_find_statement_at_depth, (void*)args,
		NULL, NULL);
  root->next = next;

  return args[2];
}

static
void traverse_past_outer_loops (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_for))
    {
      int i;
      s_past_node_t* top = ((void**)data)[0];
      s_past_node_t** outer = ((void**)data)[1];
      s_past_node_t* n;
      int is_outer = 1;
      if (node != top)
	for (n = node->parent; n; n = n->parent)
	  {
	    if (past_node_is_a (n, past_for))
	      {
		is_outer = 0;
		break;
	      }
	    if (n == top)
	      break;
	  }
      if (is_outer)
	{
	  for (i = 0; outer[i]; ++i)
	    ;
	  outer[i] = node;
	}
    }
}

static
void traverse_past_inner_loops (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_for))
    if (past_is_inner_for_loop (node))
      {
	int i;
	void** args = (void**)data;
	s_past_node_t** inner = args[1];
	for (i = 0; inner[i]; ++i)
	  ;
	inner[i] = node;
      }
}

static
s_past_node_t**
past_innerouter_loops (s_past_node_t* root, past_fun_t prefixfun)
{
  int num_for = past_count_for_loops (root);
  s_past_node_t** loops = NULL;
  if (root)
    {
      loops = XMALLOC(s_past_node_t*, num_for + 1);
      int i;
      for (i = 0; i <= num_for; ++i)
	loops[i] = NULL;
      void* args[2];
      args[0] = root;
      args[1] = loops;
      s_past_node_t* next = root->next;
      root->next = NULL;
      past_visitor (root, prefixfun, (void*)args, NULL, NULL);
      root->next = next;
      for (i = 0; loops[i]; ++i)
	;
      if (i < num_for)
	{
	  s_past_node_t** nloops = XMALLOC(s_past_node_t*, i + 1);
	  for (i = 0; loops[i]; ++i)
	    nloops[i] = loops[i];
	  nloops[i] = NULL;
	  XFREE(loops);
	  loops = nloops;
	}
    }
  else
    {
      loops = XMALLOC(s_past_node_t*, 1);
      loops[0] = NULL;
    }

  return loops;
}

/**
 * Return a NULL-terminated array of the inner loops.
 *
 */
s_past_node_t**
past_inner_loops (s_past_node_t* root)
{
  return past_innerouter_loops (root, traverse_past_inner_loops);
}


/**
 * Return a NULL-terminated array of the outer loops.
 *
 */
s_past_node_t**
past_outer_loops (s_past_node_t* root)
{
  return past_innerouter_loops (root, traverse_past_outer_loops);
}


static
void traverse_past_contain_loop (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_for))
      *((int*)data) = 1;
}

/**
 * Return true if the tree contains a loop.
 *
 */
int
past_contain_loop (s_past_node_t* root)
{
  int has_loop = 0;
  if (root)
    {
      s_past_node_t* next = root->next;
      root->next = NULL;
      past_visitor (root, traverse_past_contain_loop, &has_loop, NULL, NULL);
      root->next = next;
    }

  return has_loop;
}


/**
 * Return true if the node is an outer loop.
 *
 */
int
past_is_outer_for_loop (s_past_node_t* node)
{
  if (! past_node_is_a (node, past_for))
    return 0;
  for (node = node->parent; node; node = node->parent)
    if (past_node_is_a (node, past_for))
      return 0;

  return 1;
}


/**
 * Return true if the node is an inner loop.
 *
 */
int
past_is_inner_for_loop (s_past_node_t* node)
{
  if (! past_node_is_a (node, past_for))
    return 0;
  PAST_DECLARE_TYPED(for, pf, node);
  int num_for = 0;
  s_past_node_t* tmp;
  for (tmp = pf->body; tmp; tmp = tmp->next)
    num_for += past_count_for_loops (tmp);
  return num_for == 0;
}


static
void traverse_rebuild_symt (s_past_node_t* node, void* args)
{
  if (past_node_is_a (node, past_variable))
    {
      PAST_DECLARE_TYPED(variable, pv, node);
      s_symbol_table_t* table = args;
      if (pv->symbol->is_char_data)
	{
	  s_symbol_t* sym = symbol_add_from_char (table, pv->symbol->data);
	  if (sym != pv->symbol)
	    {
	      sym->type = pv->symbol->type;
	      sym->generated = pv->symbol->generated;
	      symbol_free (pv->symbol);
	      pv->symbol = sym;
	    }
	}
    }
  else if (past_node_is_a (node, past_for))
    {
      PAST_DECLARE_TYPED(for, pf, node);
      s_symbol_table_t* table = args;
      if (pf->iterator->symbol->is_char_data)
	{
	  s_symbol_t* sym =
	    symbol_add_from_char (table, pf->iterator->symbol->data);
	  if (sym != pf->iterator->symbol)
	    {
	      sym->type = pf->iterator->symbol->type;
	      sym->generated = pf->iterator->symbol->generated;
	      symbol_free (pf->iterator->symbol);
	      pf->iterator->symbol = sym;
	    }
	}
    }
}
/**
 * Rebuild the symbol table, and attach it to the root node.
 *
 */
void
past_rebuild_symbol_table (s_past_node_t* node)
{
  if (! past_node_is_a (node, past_root))
    return;
  PAST_DECLARE_TYPED(root, pr, node);
  if (pr->symboltable == NULL)
    pr->symboltable = symbol_table_malloc ();
  s_symbol_table_t* table = pr->symboltable;
  past_visitor (pr->body, traverse_rebuild_symt, table, NULL, NULL);
}

/**
 * Loop bound hoisting routines. See down for the entry point.
 *
 */
static
s_symbol_t* find_available_symbol (s_past_node_t* root,
				   s_past_node_t*** symbols)
{
  int i;
  char buffer[512];
  for (i = 0; symbols[i]; ++i)
    ;
  sprintf (buffer, "__loop_bound_%d", i);
  s_symbol_t* ret = symbol_add_from_char (NULL, buffer);

  return ret;
}
static
void traverse_collect_symbs (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_variable))
    {
      int i;
      s_symbol_t** syms = data;
      PAST_DECLARE_TYPED(variable, pv, node);
      for (i = 0; syms[i] && !symbol_equal (syms[i], pv->symbol); ++i)
	;
      syms[i] = pv->symbol;
    }
}
static
void process_hoist_conditional (s_past_node_t* root,
				s_past_node_t*** insert_at,
				s_past_node_t** loopexpr,
				s_past_node_t* node)
{
  // 1- Create a temporary variable.
  s_symbol_t* newvars = find_available_symbol (root, insert_at);
  s_past_variable_t* vard = past_variable_create (newvars);
  s_past_variable_t* type =
    past_variable_create (symbol_add_from_char (NULL, "int"));
  s_past_node_t* decl = past_node_vardecl_create (&(vard->node),
						  (s_past_variable_t*)type);
  s_past_node_t* newass = past_node_statement_create
    (past_node_binary_create (past_assign, decl, *loopexpr));
  s_symbol_t* s2 = symbol_add_from_char (NULL, newvars->data);
  s_past_node_t* var = past_node_variable_create (s2);
  s_past_node_t* lbexpr = *loopexpr;
  *loopexpr = var;
  // 2- Find the best place to put it.
  int nb_syms = past_count_nodetype (lbexpr, past_variable);
  s_symbol_t** syms = XMALLOC(s_symbol_t*, nb_syms + 1);
  int i;
  for (i = 0; i < nb_syms + 1; ++i)
    syms[i] = NULL;
  past_visitor (lbexpr, traverse_collect_symbs, syms, NULL, NULL);

  s_past_node_t* parent;
  s_past_node_t* before = node;
  for (parent = node->parent; parent; parent = parent->parent)
    {
      if (past_node_is_a (parent, past_for))
	{
	  PAST_DECLARE_TYPED(for, pf2, parent);
	  s_symbol_t* ls = pf2->iterator->symbol;
	  for (i = 0; syms[i] && !symbol_equal (syms[i], ls); ++i)
	    ;
	  if (! syms[i])
	    before = parent;
	  else
	    break;
	}
    }

  // 3- Store the insert_before.
  for (i = 0; insert_at[i]; ++i)
    ;
  insert_at[i] = XMALLOC(s_past_node_t*, 2);
  insert_at[i][0] = before;
  insert_at[i][1] = newass;
}
static
void traverse_is_complex_bound (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_min) ||
      past_node_is_a (node, past_max) ||
      past_node_is_a (node, past_ceil) ||
      past_node_is_a (node, past_floor) ||
      past_node_is_a (node, past_funcall) ||
      past_node_is_a (node, past_ceild) ||
      past_node_is_a (node, past_floord) ||
      past_node_is_a (node, past_round) ||
      past_node_is_a (node, past_mul) ||
      past_node_is_a (node, past_add) ||
      past_node_is_a (node, past_sub) ||
      past_node_is_a (node, past_div) ||
      past_node_is_a (node, past_sqrt))
    {
      int* has_complex = data;
      *has_complex = 1;
    }
}
static
int is_complex_bound (s_past_node_t* node)
{
  int has_complex = 0;
  past_visitor (node, traverse_is_complex_bound, &has_complex, NULL, NULL);

  return has_complex;
}
static
void traverse_optimize_loop_bound (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_for))
    {
      PAST_DECLARE_TYPED(for, pf, node);
      if (pf->init)
	{
	  PAST_DECLARE_TYPED(binary, pinit, pf->init);
	  if (! past_node_is_a (pinit->rhs, past_variable) &&
	      ! past_node_is_a (pinit->rhs, past_value))
	    {
	      // Hoist only bounds that contain min, max, floor, ceil, round.
	      if (is_complex_bound (pinit->rhs))
		{
		  void** args = (void**)data;
		  process_hoist_conditional (args[0], args[1], &(pinit->rhs),
					     node);
		}
	    }
	}
      if (pf->test)
	{
	  PAST_DECLARE_TYPED(binary, ptest, pf->test);
	  // Process only simple bounds like 'i <= blabla'.
	  if (past_node_is_a (ptest->lhs, past_variable))
	    {
	      PAST_DECLARE_TYPED(variable, pv, ptest->lhs);
	      if (! symbol_equal (pv->symbol, pf->iterator->symbol))
		return;

	      // Hoist only 'complex' loop bounds.
	      if (past_node_is_a (ptest->rhs, past_variable) ||
		  past_node_is_a (ptest->rhs, past_value))
		return;
	      if (is_complex_bound (ptest->rhs))
		{
		  void** args = (void**)data;
		  process_hoist_conditional (args[0], args[1], &(ptest->rhs),
					     node);
		}
	    }
	}
    }
}
/**
 * Optimize 'for' loop bounds, by hoisting them as much as possible.
 *
 */
void
past_optimize_loop_bounds (s_past_node_t* root)
{
  int num_for_loops = past_count_for_loops (root);
  s_past_node_t*** insert_at = XMALLOC(s_past_node_t**, 2 * num_for_loops + 1);
  int i;
  for (i = 0; i < 2 * num_for_loops + 1; ++i)
    insert_at[i] = NULL;
  void* args[2]; args[0] = root; args[1] = insert_at;

  past_visitor (root, traverse_optimize_loop_bound, (void*)args, NULL, NULL);

  past_set_parent (root);
  for (i = 0; insert_at[i]; ++i)
    {
      insert_at[i][1]->next = insert_at[i][0];
      past_replace_node (insert_at[i][0], insert_at[i][1]);
    }
}


/**
 * Return the immediate enclosing node of type 'type', NULL if it does
 * not exists.
 *
 */
s_past_node_t* past_get_enclosing_node (s_past_node_t* node,
					cs_past_node_type_t* type)
{
  while (node && node->type != type)
    node = node->parent;

  return node;
}
/**
 * FIXME: Create sub-hierarchy with write and read nodes.
 */
static int is_write_node (s_past_node_t* node)
{
  if (node->type == past_inc_before ||
      node->type == past_inc_after ||
      node->type == past_dec_before ||
      node->type == past_dec_after ||
      node->type == past_assign ||
      node->type == past_addassign ||
      node->type == past_subassign ||
      node->type == past_mulassign ||
      node->type == past_divassign ||
      node->type == past_modassign ||
      node->type == past_andassign ||
      node->type == past_orassign ||
      node->type == past_xorassign ||
      node->type == past_lshiftassign ||
      node->type == past_rshiftassign)
      return 1;
  else
      return 0;
}
/**
 * Check if a variable symbol is read or written.
 *
 */
int past_variable_is_write_reference (s_past_variable_t* var)
{
  s_past_node_t* parent = var->node.parent;
  s_past_node_t* last = &(var->node);
  while (parent && !is_write_node (parent))
    {
      if (past_node_is_a (parent, past_arrayref))
	{
	  PAST_DECLARE_TYPED(binary, pa, parent);
	  if (pa->rhs == last)
	    return 0;
	}
      last = parent;
      parent = parent->parent;
    }
  if (parent && past_node_is_a (parent, past_binary))
    {
      PAST_DECLARE_TYPED(binary, pb, parent);
      if (pb->lhs == last)
	return 1;
      return 0;
    }

  return parent != NULL;
}

static
void
traverse_past_encaps (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_arrayref))
    {
      // postfix rewriting.
      if (past_node_is_a (node->parent, past_arrayref))
	return;
      // 1- count the array depth.
      int dim = 0;
      s_past_node_t* tmp = node;
      s_past_binary_t* bin;
      while (past_node_is_a (tmp, past_arrayref))
	{
	  bin = (s_past_binary_t*)tmp;
	  tmp = bin->lhs;
	  ++dim;
	}
      // 2- Collect the index expressions, and original arrayref pointers.
      s_past_node_t* exprs[dim];
      s_past_node_t* aref[dim];
      s_past_node_t* tmp2 = node;
      int count = dim - 1;
      while (past_node_is_a (tmp2, past_arrayref))
	{
	  aref[count] = tmp2;
	  PAST_DECLARE_TYPED(binary, tp, tmp2);
	  exprs[count--] = tp->rhs;
	  tmp2 = tp->lhs;
	}
      // 3- Collect the array name.
      assert (past_node_is_a (bin->lhs, past_variable));
      s_past_variable_t* var = (s_past_variable_t*)bin->lhs;

      // 4- See if it is a read or a write reference.
      int is_write = past_variable_is_write_reference (var);

      // 5- Create the replacement node. Prototype function call:
      // __past_arrayref_{read,write}_{x}d(...)
      char call[64];
      sprintf (call, "__past_arrayref_%s_%dd",
	       is_write ? "write" : "read", dim);
      s_past_node_t* args = NULL;
      s_past_node_t* cur;
      for (count = 0; count < dim; ++count)
	if (args == NULL)
	  cur = args = exprs[count];
	else
	  cur = cur->next = exprs[count];
      s_past_variable_t* fname =
	past_variable_create (symbol_add_from_char (NULL, call));
      s_past_node_t* repl =
	past_node_funcall_create ((s_past_variable_t*)fname, args);
      if (is_write)
	repl = past_node_unary_create (past_derefof, repl);

      // 6- Replace the current node by the function call.
      past_replace_node (node, repl);

      // 7- Delete the trailing arrayref nodes.
      for (count = 0; count < dim; ++count)
	XFREE(aref[count]);
    }
}
/**
 * Replace all array references by an API-specified function call.
 *
 * E.g:
 * A[i+42][j] = ... is replaced by *__past_arrayref_write_2d(A, i+42, j) = ...
 *
 */
s_past_node_t* past_encapsulate_arrayref_in_func (s_past_node_t* root)
{
  if (past_node_is_a (root, past_arrayref))
    {
      s_past_node_t* fkroot = past_node_root_create (NULL, root);
      past_visitor (fkroot, NULL, NULL, traverse_past_encaps, NULL);
      PAST_DECLARE_TYPED(root, r, fkroot);
      root = r->body;
      XFREE(fkroot);
    }
  else if (root)
    past_visitor (root, NULL, NULL, traverse_past_encaps, NULL);

  return root;
}


static
s_symbol_t** past_collect_rw_symbols (s_past_node_t* node,
				      int read_syms, int array_syms,
				      int affineexpr_syms)
{
  int num_var = past_count_nodetype (node, past_variable);
  s_symbol_t** ret = XMALLOC(s_symbol_t*, num_var + 1);

  s_past_node_t** nodes = past_collect_nodetype (node, past_variable);
  int i, j;
  for (i = 0; i < num_var + 1; ++i)
    ret[i] = NULL;
  for (i = 0; nodes[i]; ++i)
    {
      PAST_DECLARE_TYPED(variable, var, nodes[i]);
      // Ensure the symbol is not already in the set.
      s_symbol_t* s = var->symbol;
      for (j = 0; ret[j]; ++j)
      	if (symbol_equal (ret[j], s))
      	  break;
      if (ret[j])
      	continue;

      // Check the properties of the reference.
      int is_write = past_variable_is_write_reference (var);
      int is_array = 0;
      if (nodes[i]->parent && past_node_is_a (nodes[i]->parent, past_arrayref))
      	{
      	  PAST_DECLARE_TYPED(binary, ar, nodes[i]->parent);
      	  if (ar->lhs == nodes[i])
	    is_array = 1;
      	}
      int is_affineexpr = 0;
      s_past_node_t* tmp = past_get_enclosing_node (nodes[i], past_arrayref);
      if (tmp)
	{
	  PAST_DECLARE_TYPED(binary, pb, tmp);
	  if (pb->lhs == nodes[i])
	    is_affineexpr = 0;
	  else is_affineexpr = 1;
	}
      else
	is_affineexpr = past_get_enclosing_node
	  (nodes[i], past_statement) == NULL;

      // Add the symbol if necessary.
      if (!is_write && read_syms && !array_syms && !affineexpr_syms)
	ret[j] = s;
      else if (is_write && !read_syms && !array_syms && !affineexpr_syms)
	ret[j] = s;
      else if (is_array && array_syms && !affineexpr_syms)
	ret[j] = s;
      else if (is_affineexpr && !array_syms && affineexpr_syms)
	ret[j] = s;
    }
  XFREE(nodes);

  return ret;
}

/**
 * Return a null-terminated set of unique symbols being read in any
 * affine expression in the tree. By construction those should be
 * integers.
 *
 */
s_symbol_t** past_collect_affineexpr_symbols (s_past_node_t* node)
{
  return past_collect_rw_symbols (node, 0, 0, 1);
}

/**
 * Return a null-terminated set of unique symbols being read in the
 * tree.
 *
 */
s_symbol_t** past_collect_read_symbols (s_past_node_t* node)
{
  return past_collect_rw_symbols (node, 1, 0, 0);
}

/**
 * Return a null-terminated set of unique symbols being written in the
 * tree.
 *
 */
s_symbol_t** past_collect_write_symbols (s_past_node_t* node)
{
  return past_collect_rw_symbols (node, 0, 0, 0);
}

/**
 * Return a null-terminated set of unique array symbols in the tree.
 *
 */
s_symbol_t** past_collect_array_symbols (s_past_node_t* node)
{
  return past_collect_rw_symbols (node, 0, 1, 0);
}

/**
 * Return a null-terminated set of unique loop iterator symbols
 * assigned in the tree.
 *
 */
s_symbol_t** past_collect_written_loopiterators_symbols (s_past_node_t* node)
{
  s_past_node_t** loops = past_collect_nodetype (node, past_for);
  int num_loops = past_count_nodetype (node, past_for);
  s_symbol_t** res = XMALLOC(s_symbol_t*, num_loops + 1);
  int i, j;
  for (i = 0; loops[i]; ++i)
    {
      PAST_DECLARE_TYPED(for, pf, loops[i]);
      s_symbol_t* s = pf->iterator->symbol;
      for (j = 0; j < i && res[j] != s; ++j)
	;
      res[j] = s;
    }
  XFREE(loops);

  return res;
}

static
int is_included (s_symbol_t* s, s_symbol_t** syms)
{
  int i;
  for (i = 0; syms && syms[i]; ++i)
    if (s == syms[i])
      return 1;
  return 0;
}

/**
 * Return true if the for stride is +1.
 *
 */
static
int past_for_is_stride_one (s_past_for_t* f)
{
  s_past_node_t* inc = f->increment;
  if (past_node_is_a (inc, past_inc_before) ||
      past_node_is_a(inc, past_inc_after))
    return 1;
  if (past_node_is_a (inc, past_binary))
    {
      PAST_DECLARE_TYPED(binary, pb, inc);
      if (! past_node_is_a (pb->lhs, past_variable))
	return 0;
      PAST_DECLARE_TYPED(variable, pv, pb->lhs);
      if (! symbol_equal (pv->symbol, f->iterator->symbol))
	return 0;
      if (past_node_is_a (pb->rhs, past_binary))
	{
	  if (! past_node_is_a (pb->rhs, past_add))
	    return 0;
	  PAST_DECLARE_TYPED(binary, pb2, pb->rhs);
	  if (! past_node_is_a (pb2->lhs, past_variable))
	    return 0;
	  PAST_DECLARE_TYPED(variable, pv2, pb2->lhs);
	  if (! symbol_equal (pv2->symbol, f->iterator->symbol))
	    return 0;
	  if (! past_node_is_a (pb2->rhs, past_value))
	    return 0;
	  PAST_DECLARE_TYPED(value, pvv, pb2->rhs);
	  if (pvv->value.intval != 1)
	    return 0;
	  return 1;
	}
      else
	{
	  if (! past_node_is_a (inc, past_addassign))
	    return 0;
	  if (! past_node_is_a (pb->rhs, past_value))
	    return 0;
	  PAST_DECLARE_TYPED(value, pvv, pb->rhs);
	  if (pvv->value.intval != 1)
	    return 0;
	  return 1;
	}
    }
  return 0;
}


static
void traverse_past_simdf (s_past_node_t* node, void* data)
{
  if (past_node_is_a (node, past_for))
    {
      s_past_node_t** il = past_inner_loops (node);
      if (il && il[0] == node)
	{
	  // Ensure it is a SIMD loop, or a decorated loop.
	  PAST_DECLARE_TYPED(for, pf, node);
	  /// FIXME: Temporary assume that any SIMD loop is not
	  /// stride-1 and/or decorated with some random decoration.
	  if (node->decoration == NULL && past_for_is_stride_one (pf))
	    return;

	  // Recover the data.
	  s_past_node_t** fundecls = ((void**)data)[0];
	  int* loop_cnt = ((void**)data)[1];

	  // 1- Collect all read, written and array symbols.
	  s_past_node_t* lnext = node->next;
	  node->next = NULL;
	  s_symbol_t** read = past_collect_read_symbols (node);
	  s_symbol_t** write = past_collect_write_symbols (node);
	  s_symbol_t** arrays = past_collect_array_symbols (node);
	  s_symbol_t** affineexpr = past_collect_affineexpr_symbols (node);
	  s_symbol_t** loopiters =
	    past_collect_written_loopiterators_symbols (node);
	  node->next = lnext;

	  // 2- Create the function call.
	  char call[64];
	  sprintf (call, "__past_loop_%d", (*loop_cnt)++);
	  s_past_variable_t* fname =
	    past_variable_create (symbol_add_from_char (NULL, call));
	  // Create arguments. All written non-array references are passed by
	  // address.
	  s_past_node_t* args = NULL;
	  s_past_node_t* cur;
	  int i;
	  for (i = 0; read[i]; ++i)
	    {
	      // Skip references also written.
	      if (is_included (read[i], write))
		continue;
	      //printf ("R sym: %s\n", read[i]->data);
	      s_past_node_t* vref =
		past_node_variable_create (symbol_dup (NULL, read[i]));
	      if (args == NULL)
		cur = args = vref;
	      else
		cur = cur->next = vref;
	    }
	  for (i = 0; write[i]; ++i)
	    {
	      //printf ("W sym: %s\n", write[i]->data);
	      s_past_node_t* vref =
		past_node_variable_create (symbol_dup (NULL, write[i]));
	      if (! is_included (write[i], arrays))
		vref = past_node_unary_create (past_addressof, vref);

	      if (args == NULL)
		cur = args = vref;
	      else
		cur = cur->next = vref;
	    }
	  // Create function call, and replace.
	  s_past_node_t* fcall =
	    past_node_funcall_create ((s_past_variable_t*)fname, args);
	  past_replace_node (node, fcall);
	  fcall->next = node->next;
	  node->next = NULL;

	  // 3- Create function prototype.
	  s_past_node_t* datatype = past_node_variable_create
	    (symbol_add_from_char (NULL, "DATA_TYPE"));
	  s_past_node_t* inttype = past_node_variable_create
	    (symbol_add_from_char (NULL, "int"));
	  s_past_node_t* type;
	  args = NULL;
	  for (i = 0; read[i]; ++i)
	    {
	      // Skip references also written.
	      if (is_included (read[i], write))
		continue;

	      s_past_node_t* vref = NULL;
	      // Compute the scalar type.
	      if (is_included (read[i], affineexpr))
		type = past_clone (inttype);
	      else
		type = past_clone (datatype);
	      // 3.a- Arrays.
	      if (is_included (read[i], arrays))
		{
		  // For array A[..], create 'DATA_TYPE DECL_ARRAY_A(A)'
		  vref = past_node_variable_create (symbol_dup (NULL, read[i]));
		  if (read[i]->is_char_data)
		    {
		      char macro[64];
		      sprintf (macro, "DECL_ARRAY_%s", (char*)read[i]->data);
		      s_past_variable_t* mcall = past_variable_create
			(symbol_add_from_char (NULL, macro));
		      vref = past_node_funcall_create
			((s_past_variable_t*)mcall, vref);
		    }
		  vref = past_node_vardecl_create (vref,
						   (s_past_variable_t*)type);
		}
	      // 3.b- Any other.
	      else
		{
		  vref = past_node_variable_create (symbol_dup (NULL, read[i]));
		  vref = past_node_vardecl_create (vref,
						   (s_past_variable_t*)type);
		}
	      if (args == NULL)
		cur = args = vref;
	      else
		cur = cur->next = vref;
	    }
	  for (i = 0; write[i]; ++i)
	    {
	      s_past_node_t* vref = NULL;
	      // Compute the scalar type.
	      if (is_included (write[i], affineexpr))
		type = past_clone (inttype);
	      else
		type = past_clone (datatype);
	      // 3.a- Arrays.
	      if (is_included (write[i], arrays))
		{
		  // For array A[..], create 'DATA_TYPE DECL_ARRAY_A(A)'
		  vref = past_node_variable_create
		    (symbol_dup (NULL, write[i]));
		  if (write[i]->is_char_data)
		    {
		      char macro[64];
		      sprintf (macro, "DECL_ARRAY_%s", (char*)write[i]->data);
		      s_past_node_t* mcall = past_node_variable_create
		      	(symbol_add_from_char (NULL, macro));
		      vref = past_node_funcall_create
		      	((s_past_variable_t*)mcall, vref);
		    }
		  vref = past_node_vardecl_create (vref,
						   (s_past_variable_t*)type);
		}
	      // 3.b- Any other (use local name to avoid pointers
	      // inside the loop body).
	      else
		{
		  char buff[256];
		  if (write[i]->is_char_data)
		    sprintf (buff, "__local_%s", (char*)write[i]->data);
		  else
		    sprintf (buff, "__local_%p", write[i]->data);
		  vref = past_node_variable_create
		    (symbol_add_from_char (NULL, buff));
		  vref = past_node_unary_create (past_derefof, vref);
		  vref = past_node_vardecl_create (vref,
		  				   (s_past_variable_t*)type);
		}
	      if (args == NULL)
		cur = args = vref;
	      else
		cur = cur->next = vref;
	    }
	  // Create the prototype of the definition.
	  s_past_node_t* voidtype =
	    past_node_variable_create (symbol_add_from_char (NULL, "void"));
	  s_past_fundecl_t* fundecl =
	    past_fundecl_create ((s_past_variable_t*)past_clone(&(fname->node)),
				 (s_past_variable_t*)voidtype, args, NULL);

	  // 4- Insert decl/copy statements for the written non-array
	  // symbols.
	  s_past_node_t* decls = NULL;
	  s_past_node_t* copies = NULL;
	  s_past_node_t* cur2 = NULL;
	  cur = NULL;
	  for (i = 0; write[i]; ++i)
	    {
	      if (! is_included (write[i], arrays))
		{
		  s_past_node_t* vref =
		    past_node_variable_create (symbol_dup (NULL, write[i]));
		  // Compute the scalar type.
		  s_past_node_t* type;
		  if (is_included (read[i], affineexpr))
		    type = past_clone (inttype);
		  else
		    type = past_clone (datatype);
		  vref = past_node_vardecl_create (vref,
						   (s_past_variable_t*)type);
		  char buff[256];
		  if (write[i]->is_char_data)
		    sprintf (buff, "__local_%s", (char*)write[i]->data);
		  else
		    sprintf (buff, "__local_%p", write[i]->data);
		  s_past_node_t* vref2 = past_node_variable_create
		    (symbol_add_from_char (NULL, buff));
		  vref2 = past_node_unary_create (past_derefof, vref2);
		  s_past_node_t* decl =
		    past_node_binary_create (past_assign, vref, vref2);
		  decl = past_node_statement_create (decl);
		  if (decls == NULL)
		    cur = decls = decl;
		  else
		    cur = cur->next = decl;
		  vref = past_node_variable_create
		    (symbol_dup (NULL, write[i]));
		  vref2 = past_clone (vref2);
		  s_past_node_t* copy =
		    past_node_binary_create (past_assign, vref2, vref);
		  copy = past_node_statement_create (copy);
		  if (copies == NULL)
		    cur2 = copies = copy;
		  else
		    cur2 = cur2->next = copy;
		}
	    }
	  // Finalize the function definition.
	  if (decls)
	    cur->next = node;
	  s_past_node_t* bb = past_node_block_create (decls);
	  s_past_node_t* block;
	  if (decls)
	    {
	      block = decls;
	      cur->next = node;
	    }
	  else
	    block = node;
	  node->next = copies;
	  fundecl->body = block;

	  // 5- Store the new definition in the list.
	  for (i = 0; fundecls[i]; ++i)
	    ;
	  fundecls[i] = (s_past_node_t*)fundecl;

	  // Be clean.
	  XFREE(read);
	  XFREE(write);
	  XFREE(arrays);
	  XFREE(affineexpr);
	  XFREE(loopiters);
	}
      else
	XFREE(il);
    }
}

/**
 * Replace all inner-most SIMD loop by an API-specified function call.
 *
 * E.g:
 * for (i = 0; i < N; i += 2)
 *   A[k][i] *= alpha;
 * is replaced by:
 * __past_loop_1(&i, N, k, A, alpha)
 *
 * And a new tree is created and contains:
 *
 * void __past_loop_1(int* i_ptr, int N, int k, DATA_TYPE PAST_DECL_A A,
 *                    DATA_TYPE alpha) {
 *  int i = *i_ptr;
 *  for (i = 0; i < N; i += 2)
 *     A[k][i] *= alpha;
 *  *i_ptr = i;
 * }
 *
 */
s_past_node_t* past_encapsulate_simd_loops_in_func (s_past_node_t** root)
{
  if (!root || !*root)
    return NULL;
  int num_loops = past_count_nodetype (*root, past_for);
  s_past_node_t** fundecls = XMALLOC(s_past_node_t*, num_loops + 1);
  int i;
  for (i = 0; i < num_loops + 1; ++i)
    fundecls[i] = NULL;
  int loop_counter = 0;
  void* args[2]; args[0] = fundecls; args[1] = &loop_counter;
  if (past_node_is_a (*root, past_for))
    {
      s_past_node_t* fkroot = past_node_root_create (NULL, *root);
      past_rebuild_symbol_table (fkroot);
      past_visitor (fkroot, NULL, NULL, traverse_past_simdf, (void*)args);
      past_rebuild_symbol_table (fkroot);
      PAST_DECLARE_TYPED(root, r, fkroot);
      XFREE(r->symboltable->symbols);
      XFREE(r->symboltable);
      *root = r->body;
      XFREE(fkroot);
    }
  else
    {
      past_rebuild_symbol_table (*root);
      past_visitor (*root, NULL, NULL, traverse_past_simdf, (void*)args);
      past_rebuild_symbol_table (*root);
    }

  for (i = 0; fundecls[i]; ++i)
    fundecls[i]->next = fundecls[i + 1];

  return fundecls[0];
}

/**
 * Replace all inner-most SIMD loop by an API-specified function call.
 *
 * E.g:
 * for (i = 0; i < N; i += 2)
 *   A[k][i] *= alpha;
 * is replaced by:
 * __past_loop_1(N, k, alpha, A, &i)
 *
 * And a separate file 'filename' is created and contains:
 *
 * void __past_loop_1(int N, int k, DATA_TYPE alpha, DATA_TYPE PAST_DECL_A(A),
 *                    int* __local_i) {
 *  int i = *__local_i;
 *  for (i = 0; i < N; i += 2)
 *     A[k][i] *= alpha;
 *  *__local_i = i;
 * }
 *
 */
void past_encapsulate_simd_loops_in_func_file (s_past_node_t* root,
					       char* filename)
{
  FILE* loop_file = fopen (filename, "w");
  s_past_node_t* fundecls = past_encapsulate_simd_loops_in_func (&root);
  s_past_node_t* r = past_node_root_create (NULL, fundecls);
  past_pprint (loop_file, r);
  past_deep_free (r);
  fclose (loop_file);
}
